{% extends "configforms/config_page_form.html" %}
{% load i18n %}

{% block pre_fields %}
<p>
{%  blocktrans %}
 Specify your personal API Token to allow Review Board to post your review
 request activity to I Done This.
{%  endblocktrans %}
</p>
<p>
{%  blocktrans with url="https://beta.idonethis.com/u/settings" %}
 You can find the API Token within your
 <a target="_blank" href="{{url}}">I Done This Account Settings</a>.
{%  endblocktrans %}
</p>
<p>
{%  blocktrans %}
 Your activity is only posted to I Done This teams you are a member of, if
 they are configured by the Review Board administrator. You can remove the
 token at any time to disable new posts from Review Board.
{%  endblocktrans %}
</p>
{% endblock pre_fields %}
